
from os import environ
from pathlib import Path
import typing as T

import click
import pydantic


_DEFAULT_CONFIG_DIR = Path(click.get_app_dir('instawow'))


class Config(pydantic.BaseSettings):

    addon_dir: Path
    config_dir: Path = _DEFAULT_CONFIG_DIR

    @pydantic.validator('addon_dir')
    def _prepare_addon_dir(cls, value: Path) -> Path:
        value = value.expanduser().resolve()
        if not value.is_dir():
            raise ValueError
        return value

    @pydantic.validator('config_dir')
    def _prepare_config_dir(cls, value: Path) -> Path:
        value.mkdir(exist_ok=True)
        return value


class UserConfig(Config):

    @classmethod
    def read(cls):  # -> UserConfig
        "Attempt to read the config from the default path."
        return cls(addon_dir=(_DEFAULT_CONFIG_DIR/'addon_dir.txt')
                             .read_text(encoding='utf-8'))

    def write(self) -> None:
        "Write ``self.addon_dir`` on disk."
        (self.config_dir/'addon_dir.txt').write_text(str(self.addon_dir),
                                                     encoding='utf-8')




import asyncio
import typing as T

from parsel import Selector
from pydantic.datetime_parse import parse_date
from yarl import URL

from .models import Pkg, PkgOptions
from .utils import slugify


__all__ = ('CurseResolver', 'WowiResolver', 'TukuiResolver')


class Resolver:

    origin: str
    name: str

    def __init__(self, *, manager: 'Manager') -> None:
        self.manager = manager

    def __getattr__(self, name: str) -> T.Any:
        # Delegate attribute access to the manager
        return getattr(self.manager, name)

    @classmethod
    def decompose_url(cls, url: str) -> T.Optional[T.Tuple[str, str]]:
        "Break a URL down to its component `origin` and `id`."
        raise NotImplementedError

    async def resolve(self, id_or_slug: str, *, strategy: str) -> Pkg:
        "Turn an ID or slug into a `models.Pkg`."
        raise NotImplementedError


class CurseResolver(Resolver):

    origin = 'curse'
    name = 'CurseForge'

    @classmethod
    def decompose_url(cls, url):
        url = URL(url)
        if url.host in {'wow.curseforge.com', 'www.wowace.com'} \
                and len(url.parts) > 2 \
                and url.parts[1] == 'projects':
            return (cls.origin, url.parts[2])
        elif url.host == 'www.curseforge.com' \
                and len(url.parts) > 3 \
                and url.parts[1:3] == ('wow', 'addons'):
            return (cls.origin, url.parts[3])

    async def resolve(self, id_or_slug, *, strategy):
        async with self.client.get()\
                              .get('https://wow.curseforge.com/projects/' +
                                   id_or_slug) as response:
            if response.status != 200 \
                    or response.url.host not in {'wow.curseforge.com', 'www.wowace.com'}:
                raise self.PkgNonexistent
            content = await response.text()
        content = Selector(text=content, base_url=str(response.url))
        content.root.make_links_absolute()

        if strategy == 'canonical':
            file = content.css('.cf-recentfiles > li:first-child '
                               '.cf-recentfiles-credits-wrapper')
        else:
            file = max(map(int, content.css('.cf-recentfiles abbr::attr(data-epoch)')
                                       .getall()))
            file = content.xpath(f'//*[@class = "cf-recentfiles"]'
                                 f'//abbr[@data-epoch = "{file}"]/..')

        return Pkg(origin=self.origin,
                   id=content.xpath('''\
//div[@class = "info-label" and text() = "Project ID"]/following-sibling::div/text()'''
                                    ).get().strip(),
                   slug=response.url.name,
                   name=content.css('meta[property="og:title"]::attr(content)').get(),
                   description=content.css('meta[property="og:description"]'
                                           '::attr(content)').get(),
                   url=content.css('.view-on-curse > a::attr(href)').get(),
                   file_id=URL(file.css('.overflow-tip::attr(href)').get()).name,
                   download_url=file.css('.fa-icon-download::attr(href)').get(),
                   date_published=file.css('abbr::attr(data-epoch)').get(),
                   version=file.css('.overflow-tip::attr(data-name)').get(),
                   options=PkgOptions(strategy=strategy))


class WowiResolver(Resolver):

    origin = 'wowi'
    name = 'WoWInterface'

    _api_list = 'https://api.mmoui.com/v3/game/WOW/filelist.json'
    _api_details = 'https://api.mmoui.com/v3/game/WOW/filedetails/{}.json'

    def __init__(self, *, manager):
        super().__init__(manager=manager)
        self._sync_lock = asyncio.Lock(loop=self.loop)
        self._sync_data = None

    @classmethod
    def decompose_url(cls, url):
        url = URL(url)
        if url.host in {'wowinterface.com', 'www.wowinterface.com'} \
                and len(url.parts) == 3 \
                and url.parts[1] == 'downloads':
            id_ = ''.join(c for c in url.name.split('-')[0] if c.isdigit())
            if id_:
                return (cls.origin, id_)

    async def _sync(self):
        async with self._sync_lock:
            if not self._sync_data:
                async with self.client.get()\
                                      .get(self._api_list) as response:
                    self._sync_data = {i['UID']: i for i in (await response.json())}
            return self._sync_data

    async def resolve(self, id_or_slug, *, strategy):
        try:
            addon = (await self._sync())[id_or_slug.partition('-')[0]]
        except KeyError:
            raise self.PkgNonexistent
        async with self.client.get()\
                              .get(self._api_details.format(addon['UID'])) as response:
            addon_details, = await response.json()
        if addon_details['UIPending'] == '1':
            raise self.PkgTemporarilyUnavailable

        return Pkg(origin=self.origin,
                   id=addon['UID'],
                   slug=slugify(f'{addon["UID"]} {addon["UIName"]}'),
                   name=addon['UIName'],
                   description=addon_details['UIDescription'],
                   url=addon['UIFileInfoURL'],
                   file_id=addon_details['UIMD5'],
                   download_url=addon_details['UIDownload'],
                   date_published=addon_details['UIDate'],
                   version=addon_details['UIVersion'],
                   options=PkgOptions(strategy=strategy))


class TukuiResolver(Resolver):

    origin = 'tukui'
    name = 'Tukui'

    @classmethod
    def decompose_url(cls, url):
        url = URL(url)
        if url.host == 'www.tukui.org' \
                and url.path in {'/addons.php', '/download.php'}:
            id_or_slug = url.query.get('id') or url.query.get('ui')
            if id_or_slug:
                return (cls.origin, id_or_slug)

    async def resolve(self, id_or_slug, *, strategy):
        id_or_slug = id_or_slug.partition('-')[0]
        is_ui = id_or_slug in {'elvui', 'tukui'}

        async with self.client.get()\
                              .get(f'https://www.tukui.org/api.php?'
                                   f'{"ui" if is_ui else "addon"}={id_or_slug}') as response:
            if not response.content_length:
                raise self.PkgNonexistent
            addon = await response.json(content_type='text/html')

        return Pkg(origin=self.origin,
                   id=addon['id'],
                   slug=id_or_slug if is_ui else
                        slugify(f'{addon["id"]} {addon["name"]}'),
                   name=addon['name'],
                   description=addon['small_desc'],
                   url=addon['web_url'],
                   file_id=addon['lastupdate'],
                   download_url=addon['url'],
                   date_published=parse_date(addon['lastupdate']).toordinal()
                                  if is_ui else
                                  addon['lastupdate'],
                   version=addon['version'],
                   options=PkgOptions(strategy=strategy))




from collections import namedtuple
from pathlib import Path
import re
import typing as T


class TocReader:
    """Extracts key–value pairs from TOC files."""

    Entry = namedtuple('_TocEntry', 'key value')

    def __init__(self, path: Path) -> None:
        entries = (e.lstrip('# ').partition(': ')
                   for e in path.read_text(encoding='utf-8-sig').splitlines()
                   if e.startswith('## '))
        entries = {e[0]: e[2] for e in entries}
        self.entries = entries

    def __getitem__(self, key: T.Union[str, T.Tuple[str]]) -> Entry:
        if isinstance(key, tuple):
            try:
                return next(filter(lambda i: i.value,
                                   (self.__getitem__(k) for k in key)))
            except StopIteration:
                key = key[0]
        return self.Entry(key, self.entries.get(key))


def slugify(text: str, *,
            _re_lc=re.compile(r'[^0-9a-z ]')) -> str:
    "Convert an add-on name into a lower-alphanumeric slug."
    return '-'.join(_re_lc.sub(' ', text.casefold()).split())




from datetime import datetime
from pathlib import Path

import pydantic
from sqlalchemy import (Column, String, DateTime, Enum,
                        ForeignKeyConstraint, TypeDecorator)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship


def _declarative_constructor(self, **kwargs):
    for k, v in _coerces.coercers[self.__class__]\
                        .parse_obj(kwargs):
        setattr(self, k, v)


ModelBase = declarative_base(constructor=_declarative_constructor)


class _BaseCoercer(pydantic.BaseModel):
    """The coercer is used inside the declarative constructor to type-cast
    values _prior to_ inserts.  SQLAlchemy delegates this function to
    the DB API -- see https://stackoverflow.com/a/8980982.  We need this to
    happen in advance to be able to compare values with their in-database
    counterparts.  It also saves us the trouble of having to manually
    parse dates and the like.
    """

    class Config:
        allow_extra = True
        max_anystr_length = 2 ** 32


class _coerces:

    coercers = {}

    def __new__(cls, model):
        def wrapper(coercer):
            cls.coercers[model] = coercer
            return coercer
        return wrapper


class _PathType(TypeDecorator):

    impl = String

    def process_bind_param(self, value, dialect):
        if value is not None:
            return str(value)

    def process_result_value(self, value, dialect):
        if value is not None:
            return Path(value)


class Pkg(ModelBase):

    __tablename__ = 'pkg'

    origin = Column(String, primary_key=True)
    id = Column(String, primary_key=True)
    slug = Column(String, nullable=False)
    name = Column(String, nullable=False)
    description = Column(String, nullable=False)
    url = Column(String, nullable=False)
    file_id = Column(String, nullable=False)
    download_url = Column(String, nullable=False)
    date_published = Column(DateTime, nullable=False)
    version = Column(String, nullable=False)
    folders = relationship('PkgFolder', cascade='all, delete-orphan',
                           backref='pkg')
    options = relationship('PkgOptions', cascade='all, delete-orphan',
                           uselist=False)


@_coerces(Pkg)
class _PkgCoercer(_BaseCoercer):

    origin: str
    id: str
    slug: str
    name: str
    description: str
    url: str
    file_id: str
    download_url: str
    date_published: datetime
    version: str


class PkgFolder(ModelBase):

    __tablename__ = 'pkg_folder'
    __table_args__ = (ForeignKeyConstraint(['pkg_origin', 'pkg_id'],
                                           ['pkg.origin', 'pkg.id']),)

    path = Column(_PathType, primary_key=True)
    pkg_origin = Column(String, nullable=False)
    pkg_id = Column(String, nullable=False)


@_coerces(PkgFolder)
class _PkgFolderCoercer(_BaseCoercer):

    path: Path


class PkgOptions(ModelBase):

    __tablename__ = 'pkg_options'
    __table_args__ = (ForeignKeyConstraint(['pkg_origin', 'pkg_id'],
                                           ['pkg.origin', 'pkg.id']),)

    strategy = Column(Enum('canonical', 'latest'), nullable=False)
    pkg_origin = Column(String, primary_key=True)
    pkg_id = Column(String, primary_key=True)


@_coerces(PkgOptions)
class _PkgOptionsCoercer(_BaseCoercer):

    strategy: str




class ManagerResult:

    def __call__(self):
        return self


class PkgInstalled(ManagerResult):

    def __init__(self, pkg):
        super().__init__()
        self.new_pkg = pkg


class PkgUpdated(ManagerResult):

    def __init__(self, old_pkg, new_pkg):
        super().__init__()
        self.old_pkg = old_pkg
        self.new_pkg = new_pkg


class PkgRemoved(ManagerResult):

    def __init__(self, pkg):
        super().__init__()
        self.old_pkg = pkg


class ManagerError(ManagerResult,
                   Exception):
    pass


class PkgAlreadyInstalled(ManagerError):
    pass


class PkgConflictsWithInstalled(ManagerError):

    def __init__(self, pkg):
        super().__init__()
        self.conflicting_pkg = pkg


class PkgConflictsWithPreexisting(ManagerError):

    def __init__(self, folders):
        super().__init__()
        self.folders = folders


class PkgNonexistent(ManagerError):
    pass


class PkgTemporarilyUnavailable(ManagerError):
    pass


class PkgNotInstalled(ManagerError):
    pass


class PkgOriginInvalid(ManagerError):

    def __init__(self, origin):
        super().__init__()
        self.origin = origin


class PkgUpToDate(ManagerError):
    pass


class InternalError(ManagerResult,
                    Exception):

    def __init__(self, error):
        super().__init__()
        self.error = error




from pathlib import Path
from setuptools import setup

from instawow import __version__


setup(name='instawow',
      version=__version__,
      description='A CLI for managing World of Warcraft add-ons.',
      url='http://github.com/layday/instawow',
      author='layday',
      author_email='layday@protonmail.com',
      license='GPLv3',
      long_description=Path('README.rst').read_text(),
      python_requires='~=3.6',
      packages=['instawow'],
      install_requires=['aiohttp==3.5.4',
                        'click==7.0',
                        'contextvars==2.3; python_version < "3.7"',
                        'Logbook==1.4.3',
                        'outdated==0.2.0',
                        'parsel==1.5.1',
                        'pydantic==0.18.2',
                        'Send2Trash==1.5.0',
                        'SQLAlchemy==1.2.16',
                        'texttable==1.6.0',
                        'tqdm==4.29.1',
                        'yarl==1.3.0',
                        'uvloop==0.12.0; platform_system != "Windows"',],
      package_data={'instawow': ['assets/*']},
      entry_points={'console_scripts': ['instawow = instawow.cli:cli']})
